@@ -12,16 +12,12 @@
 namespace Symfony\Component\Lock\Store;
 
 use Doctrine\DBAL\Connection;
-use Doctrine\DBAL\DBALException;
-use Doctrine\DBAL\DriverManager;
-use Doctrine\DBAL\Exception;
 use Doctrine\DBAL\Schema\Schema;
 use Symfony\Component\Lock\Exception\InvalidArgumentException;
 use Symfony\Component\Lock\Exception\InvalidTtlException;
 use Symfony\Component\Lock\Exception\LockConflictedException;
-use Symfony\Component\Lock\Exception\NotSupportedException;
 use Symfony\Component\Lock\Key;
-use Symfony\Component\Lock\StoreInterface;
+use Symfony\Component\Lock\PersistingStoreInterface;
 
 /**
  * PdoStore is a PersistingStoreInterface implementation using a PDO connection.
@@ -36,27 +32,24 @@ use Symfony\Component\Lock\StoreInterface;
  *
  * @author Jérémy Derussé <jeremy@derusse.com>
  */
-class PdoStore implements StoreInterface
+class PdoStore implements PersistingStoreInterface
 {
+    use DatabaseTableTrait;
     use ExpiringStoreTrait;
 
     private $conn;
     private $dsn;
     private $driver;
-    private $table = 'lock_keys';
-    private $idCol = 'key_id';
-    private $tokenCol = 'key_token';
-    private $expirationCol = 'key_expiration';
     private $username = '';
     private $password = '';
     private $connectionOptions = [];
-    private $gcProbability;
-    private $initialTtl;
+
+    private $dbalStore;
 
     /**
-     * You can either pass an existing database connection as PDO instance or
-     * a Doctrine DBAL Connection or a DSN string that will be used to
-     * lazy-connect to the database when the lock is actually used.
+     * You can either pass an existing database connection as PDO instance
+     * or a DSN string that will be used to lazy-connect to the database
+     * when the lock is actually used.
      *
      * List of available options:
      *  * db_table: The name of the table [default: lock_keys]
@@ -67,49 +60,41 @@ class PdoStore implements StoreInterface
      *  * db_password: The password when lazy-connect [default: '']
      *  * db_connection_options: An array of driver-specific connection options [default: []]
      *
-     * @param \PDO|Connection|string $connOrDsn     A \PDO or Connection instance or DSN string or null
-     * @param array                  $options       An associative array of options
-     * @param float                  $gcProbability Probability expressed as floating number between 0 and 1 to clean old locks
-     * @param int                    $initialTtl    The expiration delay of locks in seconds
+     * @param \PDO|string $connOrDsn     A \PDO instance or DSN string or null
+     * @param array       $options       An associative array of options
+     * @param float       $gcProbability Probability expressed as floating number between 0 and 1 to clean old locks
+     * @param int         $initialTtl    The expiration delay of locks in seconds
      *
      * @throws InvalidArgumentException When first argument is not PDO nor Connection nor string
      * @throws InvalidArgumentException When PDO error mode is not PDO::ERRMODE_EXCEPTION
-     * @throws InvalidArgumentException When namespace contains invalid characters
      * @throws InvalidArgumentException When the initial ttl is not valid
      */
     public function __construct($connOrDsn, array $options = [], float $gcProbability = 0.01, int $initialTtl = 300)
     {
-        if ($gcProbability < 0 || $gcProbability > 1) {
-            throw new InvalidArgumentException(sprintf('"%s" requires gcProbability between 0 and 1, "%f" given.', __METHOD__, $gcProbability));
-        }
-        if ($initialTtl < 1) {
-            throw new InvalidTtlException(sprintf('"%s()" expects a strictly positive TTL, "%d" given.', __METHOD__, $initialTtl));
+        if ($connOrDsn instanceof Connection || (\is_string($connOrDsn) && str_contains($connOrDsn, '://'))) {
+            trigger_deprecation('symfony/lock', '5.4', 'Usage of a DBAL Connection with "%s" is deprecated and will be removed in symfony 6.0. Use "%s" instead.', __CLASS__, DoctrineDbalStore::class);
+            $this->dbalStore = new DoctrineDbalStore($connOrDsn, $options, $gcProbability, $initialTtl);
+
+            return;
         }
 
+        $this->init($options, $gcProbability, $initialTtl);
+
         if ($connOrDsn instanceof \PDO) {
             if (\PDO::ERRMODE_EXCEPTION !== $connOrDsn->getAttribute(\PDO::ATTR_ERRMODE)) {
                 throw new InvalidArgumentException(sprintf('"%s" requires PDO error mode attribute be set to throw Exceptions (i.e. $pdo->setAttribute(PDO::ATTR_ERRMODE, PDO::ERRMODE_EXCEPTION)).', __METHOD__));
             }
 
-            $this->conn = $connOrDsn;
-        } elseif ($connOrDsn instanceof Connection) {
             $this->conn = $connOrDsn;
         } elseif (\is_string($connOrDsn)) {
             $this->dsn = $connOrDsn;
         } else {
-            throw new InvalidArgumentException(sprintf('"%s" requires PDO or Doctrine\DBAL\Connection instance or DSN string as first argument, "%s" given.', __CLASS__, \is_object($connOrDsn) ? \get_class($connOrDsn) : \gettype($connOrDsn)));
+            throw new InvalidArgumentException(sprintf('"%s" requires PDO or Doctrine\DBAL\Connection instance or DSN string as first argument, "%s" given.', __CLASS__, get_debug_type($connOrDsn)));
         }
 
-        $this->table = $options['db_table'] ?? $this->table;
-        $this->idCol = $options['db_id_col'] ?? $this->idCol;
-        $this->tokenCol = $options['db_token_col'] ?? $this->tokenCol;
-        $this->expirationCol = $options['db_expiration_col'] ?? $this->expirationCol;
         $this->username = $options['db_username'] ?? $this->username;
         $this->password = $options['db_password'] ?? $this->password;
         $this->connectionOptions = $options['db_connection_options'] ?? $this->connectionOptions;
-
-        $this->gcProbability = $gcProbability;
-        $this->initialTtl = $initialTtl;
     }
 
     /**
@@ -117,47 +102,52 @@ class PdoStore implements StoreInterface
      */
     public function save(Key $key)
     {
+        if (isset($this->dbalStore)) {
+            $this->dbalStore->save($key);
+
+            return;
+        }
+
         $key->reduceLifetime($this->initialTtl);
 
         $sql = "INSERT INTO $this->table ($this->idCol, $this->tokenCol, $this->expirationCol) VALUES (:id, :token, {$this->getCurrentTimestampStatement()} + $this->initialTtl)";
-        $stmt = $this->getConnection()->prepare($sql);
+        $conn = $this->getConnection();
+        try {
+            $stmt = $conn->prepare($sql);
+        } catch (\PDOException $e) {
+            if (!$conn->inTransaction() || \in_array($this->driver, ['pgsql', 'sqlite', 'sqlsrv'], true)) {
+                $this->createTable();
+            }
+            $stmt = $conn->prepare($sql);
+        }
 
         $stmt->bindValue(':id', $this->getHashedKey($key));
         $stmt->bindValue(':token', $this->getUniqueToken($key));
 
         try {
             $stmt->execute();
-        } catch (DBALException | Exception $e) {
-            // the lock is already acquired. It could be us. Let's try to put off.
-            $this->putOffExpiration($key, $this->initialTtl);
         } catch (\PDOException $e) {
             // the lock is already acquired. It could be us. Let's try to put off.
             $this->putOffExpiration($key, $this->initialTtl);
         }
 
-        if ($this->gcProbability > 0 && (1.0 === $this->gcProbability || (random_int(0, \PHP_INT_MAX) / \PHP_INT_MAX) <= $this->gcProbability)) {
-            $this->prune();
-        }
-
+        $this->randomlyPrune();
         $this->checkNotExpired($key);
     }
 
     /**
      * {@inheritdoc}
      */
-    public function waitAndSave(Key $key)
+    public function putOffExpiration(Key $key, float $ttl)
     {
-        @trigger_error(sprintf('%s() is deprecated since Symfony 4.4 and will be removed in Symfony 5.0.', __METHOD__), \E_USER_DEPRECATED);
-        throw new NotSupportedException(sprintf('The store "%s" does not supports blocking locks.', __METHOD__));
-    }
+        if (isset($this->dbalStore)) {
+            $this->dbalStore->putOffExpiration($key, $ttl);
+
+            return;
+        }
 
-    /**
-     * {@inheritdoc}
-     */
-    public function putOffExpiration(Key $key, $ttl)
-    {
         if ($ttl < 1) {
-            throw new InvalidTtlException(sprintf('"%s()" expects a TTL greater or equals to 1 second. Got %s.', __METHOD__, $ttl));
+            throw new InvalidTtlException(sprintf('"%s()" expects a TTL greater or equals to 1 second. Got "%s".', __METHOD__, $ttl));
         }
 
         $key->reduceLifetime($ttl);
@@ -184,6 +174,12 @@ class PdoStore implements StoreInterface
      */
     public function delete(Key $key)
     {
+        if (isset($this->dbalStore)) {
+            $this->dbalStore->delete($key);
+
+            return;
+        }
+
         $sql = "DELETE FROM $this->table WHERE $this->idCol = :id AND $this->tokenCol = :token";
         $stmt = $this->getConnection()->prepare($sql);
 
@@ -197,6 +193,10 @@ class PdoStore implements StoreInterface
      */
     public function exists(Key $key)
     {
+        if (isset($this->dbalStore)) {
+            return $this->dbalStore->exists($key);
+        }
+
         $sql = "SELECT 1 FROM $this->table WHERE $this->idCol = :id AND $this->tokenCol = :token AND $this->expirationCol > {$this->getCurrentTimestampStatement()}";
         $stmt = $this->getConnection()->prepare($sql);
 
@@ -207,39 +207,11 @@ class PdoStore implements StoreInterface
         return (bool) (\is_object($result) ? $result->fetchOne() : $stmt->fetchColumn());
     }
 
-    /**
-     * Returns a hashed version of the key.
-     */
-    private function getHashedKey(Key $key): string
-    {
-        return hash('sha256', (string) $key);
-    }
-
-    private function getUniqueToken(Key $key): string
-    {
-        if (!$key->hasState(__CLASS__)) {
-            $token = base64_encode(random_bytes(32));
-            $key->setState(__CLASS__, $token);
-        }
-
-        return $key->getState(__CLASS__);
-    }
-
-    /**
-     * @return \PDO|Connection
-     */
-    private function getConnection()
+    private function getConnection(): \PDO
     {
         if (null === $this->conn) {
-            if (strpos($this->dsn, '://')) {
-                if (!class_exists(DriverManager::class)) {
-                    throw new InvalidArgumentException(sprintf('Failed to parse the DSN "%s". Try running "composer require doctrine/dbal".', $this->dsn));
-                }
-                $this->conn = DriverManager::getConnection(['url' => $this->dsn]);
-            } else {
-                $this->conn = new \PDO($this->dsn, $this->username, $this->password, $this->connectionOptions);
-                $this->conn->setAttribute(\PDO::ATTR_ERRMODE, \PDO::ERRMODE_EXCEPTION);
-            }
+            $this->conn = new \PDO($this->dsn, $this->username, $this->password, $this->connectionOptions);
+            $this->conn->setAttribute(\PDO::ATTR_ERRMODE, \PDO::ERRMODE_EXCEPTION);
         }
 
         return $this->conn;
@@ -249,38 +221,23 @@ class PdoStore implements StoreInterface
      * Creates the table to store lock keys which can be called once for setup.
      *
      * @throws \PDOException    When the table already exists
-     * @throws DBALException    When the table already exists
-     * @throws Exception        When the table already exists
      * @throws \DomainException When an unsupported PDO driver is used
      */
     public function createTable(): void
     {
-        // connect if we are not yet
-        $conn = $this->getConnection();
-        $driver = $this->getDriver();
-
-        if ($conn instanceof Connection) {
-            $schema = new Schema();
-            $table = $schema->createTable($this->table);
-            $table->addColumn($this->idCol, 'string', ['length' => 64]);
-            $table->addColumn($this->tokenCol, 'string', ['length' => 44]);
-            $table->addColumn($this->expirationCol, 'integer', ['unsigned' => true]);
-            $table->setPrimaryKey([$this->idCol]);
-
-            foreach ($schema->toSql($conn->getDatabasePlatform()) as $sql) {
-                if (method_exists($conn, 'executeStatement')) {
-                    $conn->executeStatement($sql);
-                } else {
-                    $conn->exec($sql);
-                }
-            }
+        if (isset($this->dbalStore)) {
+            $this->dbalStore->createTable();
 
             return;
         }
 
+        // connect if we are not yet
+        $conn = $this->getConnection();
+        $driver = $this->getDriver();
+
         switch ($driver) {
             case 'mysql':
-                $sql = "CREATE TABLE $this->table ($this->idCol VARCHAR(64) NOT NULL PRIMARY KEY, $this->tokenCol VARCHAR(44) NOT NULL, $this->expirationCol INTEGER UNSIGNED NOT NULL) COLLATE utf8_bin, ENGINE = InnoDB";
+                $sql = "CREATE TABLE $this->table ($this->idCol VARCHAR(64) NOT NULL PRIMARY KEY, $this->tokenCol VARCHAR(44) NOT NULL, $this->expirationCol INTEGER UNSIGNED NOT NULL) COLLATE utf8mb4_bin, ENGINE = InnoDB";
                 break;
             case 'sqlite':
                 $sql = "CREATE TABLE $this->table ($this->idCol TEXT NOT NULL PRIMARY KEY, $this->tokenCol TEXT NOT NULL, $this->expirationCol INTEGER)";
@@ -295,14 +252,26 @@ class PdoStore implements StoreInterface
                 $sql = "CREATE TABLE $this->table ($this->idCol VARCHAR(64) NOT NULL PRIMARY KEY, $this->tokenCol VARCHAR(64) NOT NULL, $this->expirationCol INTEGER)";
                 break;
             default:
-                throw new \DomainException(sprintf('Creating the lock table is currently not implemented for PDO driver "%s".', $driver));
+                throw new \DomainException(sprintf('Creating the lock table is currently not implemented for platform "%s".', $driver));
         }
 
-        if (method_exists($conn, 'executeStatement')) {
-            $conn->executeStatement($sql);
-        } else {
-            $conn->exec($sql);
+        $conn->exec($sql);
+    }
+
+    /**
+     * Adds the Table to the Schema if it doesn't exist.
+     *
+     * @deprecated since symfony/lock 5.4 use DoctrineDbalStore instead
+     */
+    public function configureSchema(Schema $schema): void
+    {
+        if (isset($this->dbalStore)) {
+            $this->dbalStore->configureSchema($schema);
+
+            return;
         }
+
+        throw new \BadMethodCallException(sprintf('"%s::%s()" is only supported when using a doctrine/dbal Connection.', __CLASS__, __METHOD__));
     }
 
     /**
@@ -312,12 +281,7 @@ class PdoStore implements StoreInterface
     {
         $sql = "DELETE FROM $this->table WHERE $this->expirationCol <= {$this->getCurrentTimestampStatement()}";
 
-        $conn = $this->getConnection();
-        if (method_exists($conn, 'executeStatement')) {
-            $conn->executeStatement($sql);
-        } else {
-            $conn->exec($sql);
-        }
+        $this->getConnection()->exec($sql);
     }
 
     private function getDriver(): string
@@ -326,47 +290,14 @@ class PdoStore implements StoreInterface
             return $this->driver;
         }
 
-        $con = $this->getConnection();
-        if ($con instanceof \PDO) {
-            $this->driver = $con->getAttribute(\PDO::ATTR_DRIVER_NAME);
-        } else {
-            $driver = $con->getDriver();
-
-            switch (true) {
-                case $driver instanceof \Doctrine\DBAL\Driver\Mysqli\Driver:
-                    throw new \LogicException(sprintf('The adapter "%s" does not support the mysqli driver, use pdo_mysql instead.', static::class));
-                case $driver instanceof \Doctrine\DBAL\Driver\AbstractMySQLDriver:
-                    $this->driver = 'mysql';
-                    break;
-                case $driver instanceof \Doctrine\DBAL\Driver\PDOSqlite\Driver:
-                case $driver instanceof \Doctrine\DBAL\Driver\PDO\SQLite\Driver:
-                    $this->driver = 'sqlite';
-                    break;
-                case $driver instanceof \Doctrine\DBAL\Driver\PDOPgSql\Driver:
-                case $driver instanceof \Doctrine\DBAL\Driver\PDO\PgSQL\Driver:
-                    $this->driver = 'pgsql';
-                    break;
-                case $driver instanceof \Doctrine\DBAL\Driver\OCI8\Driver:
-                case $driver instanceof \Doctrine\DBAL\Driver\PDOOracle\Driver:
-                case $driver instanceof \Doctrine\DBAL\Driver\PDO\OCI\Driver:
-                    $this->driver = 'oci';
-                    break;
-                case $driver instanceof \Doctrine\DBAL\Driver\SQLSrv\Driver:
-                case $driver instanceof \Doctrine\DBAL\Driver\PDOSqlsrv\Driver:
-                case $driver instanceof \Doctrine\DBAL\Driver\PDO\SQLSrv\Driver:
-                    $this->driver = 'sqlsrv';
-                    break;
-                default:
-                    $this->driver = \get_class($driver);
-                    break;
-            }
-        }
+        $conn = $this->getConnection();
+        $this->driver = $conn->getAttribute(\PDO::ATTR_DRIVER_NAME);
 
         return $this->driver;
     }
 
     /**
-     * Provides a SQL function to get the current timestamp regarding the current connection's driver.
+     * Provides an SQL function to get the current timestamp regarding the current connection's driver.
      */
     private function getCurrentTimestampStatement(): string
     {
@@ -382,7 +313,7 @@ class PdoStore implements StoreInterface
             case 'sqlsrv':
                 return 'DATEDIFF(s, \'1970-01-01\', GETUTCDATE())';
             default:
-                return time();
+                return (string) time();
         }
     }
 }
